from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace
from duckduckgo_search import DDGS

class Search(ComponentABC):
    def __init__(self, agent):
        self.agent = agent
        self.is_debug = self.agent.settings.get_trace_back("is_debug")
        self.settings = RuntimeCtxNamespace("plugin_settings.agent_component.Search", self.agent.settings)

    def __exec_search(self, keywords: str, options:dict={}):
        result = []
        search_kwargs = {}
        proxy = self.agent.settings.get_trace_back("proxy")
        if proxy:
            search_kwargs["proxies"] = proxy
        #else:
            #if self.is_debug:
                #print("[No Proxy for Search] That may cause http request error in some areas. If request error occured, please use .set_proxy('<Your Proxy Address>') or .set_settings('proxy', '<Your Proxy Address>') to set a proxy.")
        if "max_results" not in options:
            options.update({ "max_results": self.settings.get_trace_back("max_result_count", 5) })
        with DDGS(**search_kwargs) as ddgs:
            for r in ddgs.text(
                keywords,
                **options
            ):
                result.append(r)
        return result

    def use_search(self):
        self.agent.toggle_component("Search", True)
        return self.agent

    def set_max_results(self, max_result_count: int):
        self.settings.set("max_result_count", max_result_count)
        return self.agent

    def set_max_result_length(self, max_result_length: int):
        self.settings.set("max_result_length", max_result_length)
        return self.agent

    def switch_mode(self, mode: int):
        if mode in (0, 1):
            self.mode = mode
        return self.agent

    def _prefix(self):
        request = self.agent.worker_request
        if self.is_debug:
            print("[Agent Component] Searching: Start search judgement...")
        result = request\
            .input(self.agent.request.request_runtime_ctx.get("prompt")["input"])\
            .output({
                "certain_of_direct_answer": ("Boolean", "if you are cretain that you can answer {input} directly, return true, else false."),
                "keywords": (
                    "String",
                    "- split keywords by space.\n- if do not have one remains {keywords} as ""\n- if you want to search a definition, use the format '{keywords}' to make sure the exact search\n - if you want to search multiple times, split each search by '|'"
                ),
            })\
            .start()
        if not result["certain_of_direct_answer"]:
            search_result_dict = {}
            max_result_length = self.settings.get("max_result_length", 1000)
            result_length = 0
            keywords_list = result["keywords"].split("|")
            for keywords in keywords_list:
                if self.is_debug:
                    print(f"[Agent Component] Searching: { keywords }")
                search_result_dict.update({ keywords: [] })
                for search_result in self.__exec_search(keywords):
                    search_result = f"{ search_result['title'] }: { search_result['body'] }"
                    search_result_dict[keywords].append(search_result)
                    result_length += len(search_result)
                    if result_length > max_result_length:
                        break
                if self.is_debug:
                    print(f"[Agent Component] Search Result: { keywords }\n{ search_result_dict[keywords] }")
            if len(search_result_dict.keys()) > 0:
                return { "information": { "search results": search_result_dict }, "instruction": "answer according {information}" }
            else:
                return None
        else:
            return None

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "use_search": { "func": self.use_search },
            },
        }

def export():
    return ("Search", Search)


from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace

class Status(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.status_runtime_ctx = RuntimeCtxNamespace("status", self.agent.agent_runtime_ctx)
        self.status_mapping_runtime_ctx = RuntimeCtxNamespace("status_mapping", self.agent.agent_runtime_ctx)
        self.status_storage = self.agent.global_storage.table("status")
        self.settings = RuntimeCtxNamespace("plugin_settings.agent_component.Status", self.agent.settings)
        self.global_status_namespace = None
        if self.settings.get_trace_back("auto"):
            self.load()

    def set(self, key: str, value: str):
        self.status_runtime_ctx.set(key, value)
        return self.agent

    def use_global_status(self, namespace_name: str="default"):
        self.global_status_namespace = namespace_name
        return self.agent

    def append_mapping(self, status_key: str, status_value: str, alias_name: str, *args, **kwargs):
        self.status_mapping_runtime_ctx.append(
            f"{ status_key }.{ status_value }",
            {
                "alias_name": alias_name,
                "args": args,
                "kwargs": kwargs,
            }
        )
        return self.agent

    def save(self):
        status = self.status_runtime_ctx.get()
        self.status_storage.set(status).save()
        return self.agent

    def load(self):
        status = self.status_storage.get()
        self.status_runtime_ctx.update(status)
        return self.agent

    def _prefix(self):
        agent_status = self.status_runtime_ctx.get()
        if agent_status:
            # get mappings
            global_status_mappings_dict = self.agent.global_storage.table(f"status_mapping.{ self.global_status_namespace }").get() if self.global_status_namespace != None else {}
            agent_status_mappings_dict = self.status_mapping_runtime_ctx.get()
            if agent_status_mappings_dict == None:
                agent_status_mappings_dict = {}
            for status_key, status_value in self.status_runtime_ctx.get().items():
                # handle global mappings first
                if status_key in global_status_mappings_dict and status_value in global_status_mappings_dict[status_key]:
                    for global_status_mapping in global_status_mappings_dict[status_key][status_value]:
                        getattr(self.agent, global_status_mapping["alias_name"])(*global_status_mapping["args"], **global_status_mapping["kwargs"])
                # handle agent mappings
                if status_key in agent_status_mappings_dict and status_value in agent_status_mappings_dict[status_key]:
                    for agent_status_mapping in agent_status_mappings_dict[status_key][status_value]:
                        getattr(self.agent, agent_status_mapping["alias_name"])(*agent_status_mapping["args"], **agent_status_mapping["kwargs"])
        return None
        
    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "set_status": { "func": self.set },
                "save_status": { "func": self.save },
                "load_status": { "func": self.load },
                "use_global_status": { "func": self.use_global_status },
                "append_status_mapping": { "func": self.append_mapping },
            },
        }

def export():
    return ("Status", Status)


import uuid
import asyncio
import inspect
from ..utils import RuntimeCtx
from .utils.exec_tree import disable_chunk_dep_ticket, create_new_chunk_slot_with_val
from .utils.find import find_by_attr
from .lib.BreakingHub import BreakingHub
from .lib.Store import Store
from .lib.constants import WORKFLOW_START_DATA_HANDLE_NAME, WORKFLOW_END_DATA_HANDLE_NAME, DEFAULT_INPUT_HANDLE_VALUE, DEFAULT_OUTPUT_HANDLE_VALUE, BUILT_IN_EXECUTOR_TYPES, EXECUTOR_TYPE_CONDITION

class MainExecutor:
    def __init__(self, workflow_id, settings: RuntimeCtx, logger):
        # == Step 1. 初始化设定配置 ==
        self.workflow_id = workflow_id
        self.settings = settings
        self.max_execution_limit = self.settings.get('max_execution_limit') or 10
        self.logger = logger
        # == Step 2. 初始化状态配置 ==
        self.running_status = 'idle'
        # 中断器
        self.breaking_hub = BreakingHub(
            breaking_handler = self._handle_breaking,
            max_execution_limit=self.max_execution_limit
        )
        # 运行时数据存储
        self.store = self.settings.get('store') or Store()
        # 运行时系统存储（如存储整个 workflow 的输入输出数据）
        self.sys_store  = self.settings.get('sys_store') or Store()
        # 是否保留执行状态
        self.persist_state = self.settings.get('persist_state') == True
        self.persist_sys_state = self.settings.get('persist_sys_state') == True
        # 已注册的执行器
        self.registed_executors = {}
        # 执行节点字典
        self.chunks_map = {}

    async def start(self, executed_schema: dict, start_data: any = None, *, storage: dict = None):
        self.reset_all_runtime_status()
        # Set Initial Storage
        if storage and not isinstance(storage, dict):
            raise Exception(f"Initial storage can only be a dictionary.\nstorage = { storage }")
        if storage and isinstance(storage, dict):
            self.store.update_by_dict(storage)
        # 尝试灌入初始数据
        self.sys_store.set(WORKFLOW_START_DATA_HANDLE_NAME, start_data)
        self.chunks_map = executed_schema.get('chunk_map') or {}
        self.running_status = 'start'
        await self._execute_main(executed_schema.get('entries') or [])
        self.running_status = 'end'
        return self.sys_store.get(WORKFLOW_END_DATA_HANDLE_NAME) or None

    def regist_executor(self, name: str, executor):
        """
        注册执行器，传入执行器的名称及executor
        """
        self.registed_executors[name] = executor

    def unregist_executor(self, name: str):
        """
        取消注册执行器，传入执行器的名称
        """
        if name in self.registed_executors:
            del self.registed_executors[name]

        return self

    def reset_all_runtime_status(self):
        """重置状态配置"""
        self.running_status = 'idle'
        # 中断器
        self.breaking_hub = BreakingHub(
            breaking_handler=self._handle_breaking,
            max_execution_limit=self.max_execution_limit
        )
        # 运行时数据存储
        if not self.persist_state:
            self.store.remove_all()
        if not self.persist_sys_state:
            self.sys_store.remove_all()
        # 执行节点字典
        self.chunks_map = {}
    
    async def _execute_main(self, entries: list):
        """执行入口"""

        # 1、声明单个执行逻辑（异步）
        async def execute_from_entry(entry):
            slow_tasks = []
            await self._execute_partial(
                entry,
                slow_tasks=slow_tasks,
                executing_ids=[],
                visited_record=[]
            )
            # 如果执行完成后，还有待执行任务，则需手动执行
            if len(slow_tasks) > 0:
                await self._execute_slow_tasks(slow_tasks)

        # 2、收集执行任务
        entry_tasks = [execute_from_entry(entry_chunk) for entry_chunk in entries]

        # 3、最后再统一执行
        await asyncio.gather(*entry_tasks)

    async def _execute_partial(
            self,
            chunk,
            slow_tasks: list, # 缓执行任务
            executing_ids: list, # 执行中的任务
            visited_record: list # 已执行记录
        ):
        """一组分组执行的核心方法"""

        # 1、执行当前 chunk
        has_been_executed = await self._execute_single_chunk(
            chunk=chunk,
            slow_tasks=slow_tasks,
            executing_ids=executing_ids,
            visited_record=visited_record
        )
        # 如果根本未执行过（如无执行票据），直接返回
        if not has_been_executed:
            return

        # 2、执行下游直接子 chunk 自身
        executed_child_chunks = []
        # 2.1 声明执行处理逻辑
        async def execute_child_chunk(next_chunk):
            self.logger.debug(
                f"Try to call next chunk '{self._get_chunk_title(next_chunk)}' from chunk '{self._get_chunk_title(chunk)}'")
            child_executed = await self._execute_single_chunk(
                chunk=next_chunk,
                slow_tasks=slow_tasks,
                executing_ids=executing_ids,
                visited_record=visited_record
            )
            # 成功执行的塞入记录中（用于稍后处理其后续的chunk）
            if child_executed:
                executed_child_chunks.append(next_chunk['id'])

        # 2.2 执行直接子任务
        for next_info in chunk['next_chunks']:
            next_chunk = self.chunks_map.get(next_info['id'])
            if not next_chunk:
                continue

            # 在保持状态正确的前提下执行
            await execute_child_chunk(next_chunk)

        # 3、递归处理已执行了的下游直接子节点的下级节点
        # 3.1 收集直接子节点的待执行的下游任务
        next_child_tasks = []
        for next_child_id in executed_child_chunks:
            next_chunk = self.chunks_map.get(next_child_id)
            # 找不到定义，或者压根之前未成功执行时，则直接
            if not next_chunk:
                continue

            grandchild_chunks = next_chunk.get('next_chunks') or []
            for grandchild_chunk_info in grandchild_chunks:
                grandchild_chunk = self.chunks_map.get(
                    grandchild_chunk_info['id'])
                if not grandchild_chunk:
                    continue
                # 回到总入口，递归处理后辈节点（收集阶段，执行在后头并行执行）
                next_child_tasks.append(
                    self._execute_partial(
                        chunk=grandchild_chunk,
                        slow_tasks=slow_tasks,
                        executing_ids=executing_ids,
                        visited_record=visited_record
                    )
                )

        # 3.2 集中等待后辈任务执行完
        if len(next_child_tasks):
            await asyncio.gather(*next_child_tasks)

        # 4、最后尝试执行缓执行任务（如循环，要在慢于常规任务的执行）
        if len(executing_ids) == 0:
            await self._execute_slow_tasks(slow_tasks)
            slow_tasks.clear()
    
    async def _execute_single_chunk(
        self,
        chunk,
        slow_tasks: list,  # 缓执行任务
        executing_ids: list,  # 执行中的任务
        visited_record: list  # 已执行记录
    ):
        """执行完一个 chunk 自身（包含所有可用的依赖数据的组合）"""
        has_been_executed = False
        # 针对循环，不在本执行组内执行，存入缓执行组中，延后执行
        if (chunk.get('loop_entry') == True) and (chunk['id'] in visited_record):
            slow_tasks.append(chunk)
            self.logger.debug(
                f"Put the loop starting chunk '{self._get_chunk_title(chunk)}' into the slow tasks queue for delayed execution")
            return has_been_executed

        # 获取执行chunk的依赖数据（每个手柄可能有多份就绪的数据）
        single_dep_map = self._extract_execution_single_dep_data(chunk)
        while (single_dep_map['is_ready'] and single_dep_map['has_ticket']):
            has_been_executed = True
            # 基于依赖数据快照，执行分组
            await self._execute_single_chunk_core(
                chunk=chunk,
                executing_ids=executing_ids,
                visited_record=visited_record,
                single_dep_map=single_dep_map['data']
            )
            # 执行后，收回本次用到的执行票据（仅在自身完了后才执行）
            self._disable_dep_execution_ticket(
                single_dep_map=single_dep_map['data'],
                chunk=chunk
            )
            # 再次更新获取依赖（如没有了，则停止了）
            single_dep_map = self._extract_execution_single_dep_data(chunk)
        return has_been_executed

    async def _execute_single_chunk_core(
        self,
        chunk,
        executing_ids: list,  # 执行中的任务
        visited_record: list,  # 已执行记录
        single_dep_map: dict  # 依赖项（已拆组之后的）
    ):
        """根据某一份指定的的依赖数据，执行当前 chunk 自身（不包含下游 chunk 的执行调用）"""
        # 1、执行当前 chunk
        execute_id = str(uuid.uuid4())
        executing_ids.append(execute_id)
        # self.logger.debug("With dependent data: ", single_dep_map)
        exec_res = await self._exec_chunk_with_dep_core(chunk, single_dep_map)
        exec_value = exec_res['value']
        condition_signal = exec_res['signal']

        self.breaking_hub.recoder(chunk)  # 更新中断器信息
        visited_record.append(chunk['id'])  # 更新执行记录

        # 2、执行完成后，提取当前执行的结果，尝试将当前执行结果注入到下游的运行依赖插槽上 next_chunk['deps'][]['data_slots'][] = 'xxx'
        for next_info in chunk['next_chunks']:
            next_chunk = self.chunks_map.get(next_info['id'])
            if not next_chunk:
                continue

            # 针对目标 chunk 的目标 handle 的插槽位，注入最新的值（带执行票据的）。（注意如果是条件连接线，需要在条件满足时才更新）
            for next_rel_handle in next_info['handles']:
                source_handle = next_rel_handle['source_handle']
                target_handle = next_rel_handle['handle']
                source_value = None
                # 以下情况直接将完整值注入下游对应的插槽位置（此处的 target_handle）：执行结果为非 dict 类型，或者未定义上游 chunk 的输出句柄(此处的 source_handle)，或其输出句柄为默认全量输出句柄时
                if (not isinstance(exec_value, dict)) or (not source_handle) or (source_handle == DEFAULT_OUTPUT_HANDLE_VALUE):
                    source_value = exec_value
                else:
                    source_value = exec_value.get(source_handle)

                # 有条件的情况下，仅在条件满足时，才更新下游节点的数据
                condition_call = next_rel_handle.get('condition')
                if condition_call:
                    judge_res = condition_call(condition_signal, self.store)
                    connection_status = judge_res == True
                    self.logger.debug(
                        f"The connection status of '{self._get_chunk_title(chunk)}({source_handle})' to '{self._get_chunk_title(next_chunk)}({target_handle})': {connection_status}")
                    if not connection_status:
                        continue

                # 在下一个 chunk 的依赖定义中，找到与当前 chunk 的当前 handle 定义的部分，尝试更新其插槽值依赖
                self.logger.debug(
                    f"Try update chunk '{self._get_chunk_title(next_chunk)}' dep handle '{target_handle}' with value:{source_value}")
                next_chunk_target_dep = find_by_attr(
                    next_chunk['deps'], 'handle', target_handle)
                if next_chunk_target_dep:
                    next_chunk_dep_slots = next_chunk_target_dep['data_slots'] or []
                    # 1、首先清空掉之前由当前节点设置，但票据已失效的值
                    next_chunk_target_dep['data_slots'] = next_chunk_dep_slots = [
                        slot for slot in next_chunk_dep_slots if not ((slot['updator'] == chunk['id']) and slot['execution_ticket'] == '')
                    ]

                    # 2、再把本次新的值加入到该下游 chunk 的对应输入点的插槽位中
                    next_chunk_dep_slots.append(
                        create_new_chunk_slot_with_val(chunk['id'], source_value))

        # 任务执行完后，清理执行中的状态
        executing_ids.remove(execute_id)

    async def _execute_slow_tasks(self, slow_tasks):
        """尝试执行低优任务"""
        if len(slow_tasks) == 0:
            return

        self.logger.debug(f'Try to execute the slow tasks queue(length: {len(slow_tasks)})')
        for chunk in slow_tasks:
            # 先清空下游所有节点数据
            self._chunks_clean_walker(chunk)
            # 再启动执行
            await self._execute_partial(
                chunk,
                slow_tasks=[],
                executing_ids=[],
                visited_record=[]
            )
    
    async def _exec_chunk_with_dep_core(self, chunk, specified_deps = {}):
        """ 执行任务（执行到此处的都是上游数据已就绪了的） """
        # 简化参数
        deps_dict = {}
        for dep_handle in specified_deps:
            deps_dict[dep_handle] = specified_deps[dep_handle]['value']
        
        input_value = deps_dict

        # 激进模式下的特殊处理
        if self.settings.get('mode') == 'aggressive':
            # 如果只有一个数据挂载，且为 default，则直接取出来作为默认值
            all_keys = list(deps_dict.keys())
            if len(all_keys) == 1 and all_keys[0] == DEFAULT_INPUT_HANDLE_VALUE:
                input_value = deps_dict['default']

        # 交给执行器执行
        executor_type = chunk['data']['type']
        chunk_executor = self._get_chunk_executor(executor_type) or chunk.get('executor')
        # 是否内置的执行器（会追加系统信息）
        is_built_in_type = executor_type in BUILT_IN_EXECUTOR_TYPES
        exec_res = None
        if not chunk_executor:
            err_msg = f"Node {executor_type} Error-'{self._get_chunk_title(chunk)}'({chunk['id']}): The 'executor' is required but get 'NoneType'"
            self.logger.error(err_msg)
            # 主动中断执行
            raise Exception(err_msg)
        try:
            self.logger.info(f"Executing chunk '{self._get_chunk_title(chunk)}'")
            # 如果执行器是异步的，采用 await调用
            if inspect.iscoroutinefunction(chunk_executor):
                if is_built_in_type:
                    exec_res = await chunk_executor(input_value, self.store, sys_store=self.sys_store, chunk=chunk)
                else:
                    exec_res = await chunk_executor(input_value, self.store)
            else:
                if is_built_in_type:
                    exec_res = chunk_executor(input_value, self.store, sys_store=self.sys_store, chunk=chunk)
                else:
                    exec_res = chunk_executor(input_value, self.store)
        except Exception as e:
            self.logger.error(f"Node Execution Exception-'{self._get_chunk_title(chunk)}'({chunk['id']}):\n {e}")
            # 主动中断执行
            raise Exception(e)

        exec_value = exec_res
        condition_signal = None
        # 条件节点，拆分为条件信号和执行结果两部分
        if chunk['type'] == EXECUTOR_TYPE_CONDITION:
            exec_value = exec_res.get('values')
            condition_signal = exec_res.get('condition_signal')
        return {
            "value": exec_value,
            "signal": condition_signal
        }
    
    def _extract_execution_single_dep_data(self, chunk):
        """实时获取某个 chunk 的一组全量可执行数据（如没有，则返回 None）"""
        deps = chunk.get('deps')
        if not deps or len(deps) == 0:
            return {"is_ready": True, "data": None, "has_ticket": True}
        single_dep_map = {}
        exist_exec_ticket = False

        for dep in deps:
            slots = dep['data_slots'] or []
            handle_name = dep['handle']
            # 暂存的就绪的数据（循环中会不停更新，后头的会覆盖前头的），注意，就绪的数据也有可能就是 None，所以 None 不代表没有就绪数据
            tmp_ready_slot = None
            # 是否有就绪的值
            has_ready_value = False

            for slot in slots:
                # 找到就绪的数据
                if slot['is_ready']:
                    has_ready_value = True
                    # 先暂存就绪的数据作为临时数据（后面的会覆盖前头的）
                    tmp_ready_slot = slot

                    # 如果目前全局还没遇到过有票据，且当前为有票据的情况，则作为本次执行的消耗票据
                    if slot['execution_ticket'] and not exist_exec_ticket:
                        exist_exec_ticket = True
                        single_dep_map[handle_name] = slot
                        break

            # 如果跑完所有，还是没设置值（可能没有就绪的数据，或者就绪的数据都是有票据的），需要从就绪的数据中强行取一个作为本次的值
            if (handle_name not in single_dep_map) and has_ready_value:
                single_dep_map[handle_name] = tmp_ready_slot

            # 如果本轮跑完都没有设置值，则标识该 handle 数据未就绪，直接返回
            if handle_name not in single_dep_map:
                return {"is_ready": False, "data": None, "has_ticket": exist_exec_ticket}

        return {"is_ready": True, "data": single_dep_map, "has_ticket": exist_exec_ticket}

    def _disable_dep_execution_ticket(self, single_dep_map, chunk):
        """销毁chunk对应的依赖执行票据（一般在执行结束后操作）"""
        for dep_handle in single_dep_map:
            # 找到被执行的 id
            effect_id = single_dep_map[dep_handle]['id']
            for dep in chunk['deps']:
                slots = dep['data_slots'] or []
                for slot in slots:
                    # 将当前 chunk 中对应手柄的数据项中的对应目标数据的执行票据收回
                    if slot['id'] == effect_id:
                        disable_chunk_dep_ticket(slot)

    def _get_chunk_executor(self, name: str):
        """
        根据类型名称获取执行器
        """
        return self.registed_executors.get(name)

    def _chunks_clean_walker(self, root_chunk):
        """尝试对某个节点以下的分支做一轮清理工作"""

        visited_record = []
        def clean_core(chunk):
            for next_info in chunk['next_chunks']:
                next_chunk = self.chunks_map.get(next_info['id'])
                if not next_chunk:
                    continue

                # 同一个发起方的清理，只执行一次，避免死循环
                visited_symbol = f"{chunk['id']}-2-{next_chunk['id']}"
                if (visited_symbol in visited_record) or (next_chunk['id'] == root_chunk['id']):
                    continue

                visited_record.append(visited_symbol)
                effect_handles = [handle_desc['handle'] for handle_desc in next_info['handles']]
                for dep in next_chunk['deps']:
                    data_slots = dep['data_slots']
                    if len(data_slots):
                        for i in range(len(data_slots) - 1, -1, -1):
                            data_slot = data_slots[i]
                            # 如果当前的输入句柄是受上游影响的，则清理掉
                            if (dep['handle'] in effect_handles) and (data_slot['updator'] == '' or data_slot['updator'] == chunk['id']):
                                del data_slots[i]
                clean_core(next_chunk)

        clean_core(root_chunk)
        
    
    def _get_chunk_title(self, chunk):
        return chunk["data"]["title"] or f'chunk-{chunk["data"]["id"]}' or 'Unknow chunk'

    def _handle_breaking(self, chunk, type):
        """处理中断"""
        self.logger.error(
            f"Exceeded maximum execution limit: {self._get_chunk_title(chunk)}")
        # 中断之前处理相关逻辑
        # 主动中断执行
        raise Exception(
            f"Exceeded maximum execution limit: {self._get_chunk_title(chunk)}")



import json
import types
from tencentcloud.common import credential
from tencentcloud.common.profile.client_profile import ClientProfile
from tencentcloud.common.profile.http_profile import HttpProfile
from tencentcloud.hunyuan.v20230901 import hunyuan_client, models

from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, find_json, format_request_messages
from Agently.utils import RuntimeCtxNamespace

class Hunyuan(RequestABC):
    def __init__(self, request):
        self.request = request
        self.model_name = 'Hunyuan'
        self.model_settings = RuntimeCtxNamespace(f"model.{self.model_name}", self.request.settings)
        self.request_type = self.request.request_runtime_ctx.get('request_type', 'chat')
        self.default_options = {
            "model": "hunyuan-standard",
            "stream": True,
        }
        if not self.model_settings.get_trace_back("message_rules.no_multi_system_messages"):
            self.model_settings.set("message_rules.no_multi_system_messages", True)
        if not self.model_settings.get_trace_back("message_rules.strict_orders"):
            self.model_settings.set("message_rules.strict_orders", True)
        if not self.model_settings.get_trace_back("message_rules.no_multi_type_messages"):
            self.model_settings.set("message_rules.no_multi_type_messages", True)
    
    def construct_request_messages(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[重要指导说明]\n{ to_instruction(general_instruction_data) }" }] })
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[角色及行为设定]\n{ to_instruction(role_data) }" }] })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[关于和你对话的用户]\n{ to_instruction(user_info_data) }" }] })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "assistant", "content": [{"type": "text", "text": f"[主题及摘要]\n{ to_instruction(headline_data) }" }] })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        prompt_text = ""
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            prompt_text = to_instruction(prompt_input_data)
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[INFO]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[OUTPUT要求]"] = {
                        "TYPE": "可被解析的JSON字符串",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[OUTPUT要求]"] = str(prompt_output_data)
            prompt_text = to_prompt_structure(prompt_dict, end="[OUTPUT]:\n")
        request_messages.append({ "role": "user", "content": [{"type": "text", "text": prompt_text }] })
        return request_messages

    def _uppercase_keys(self, target_dict: dict):
        if not isinstance(target_dict, dict):
            return target_dict
        uppercased_dict = {}
        for key, value in target_dict.items():
            if isinstance(value, dict):
                value = self._uppercase_keys(value)
            uppercased_dict.update({ f"{ key[:1].upper() }{ key[1:] }": value })
        return uppercased_dict

    def _uppercase_messages_keys(self, target_messages: list):
        uppercased_messages = []
        for item in target_messages:
            uppercased_item = {}
            for key, value in item.items():
                if isinstance(value, dict):
                    value = self._uppercase_keys(value)
                uppercased_item.update({ f"{ key[:1].upper() }{ key[1:] }": value })
            uppercased_messages.append(uppercased_item)
        return uppercased_messages

    def generate_request_data(self):
        options = self.default_options.copy()
        options.update(self.model_settings.get_trace_back("options", {}))
        return {
            "Messages": self._uppercase_messages_keys(
                format_request_messages( self.construct_request_messages(), self.model_settings )
            ),
            **self._uppercase_keys(options)
        }

    async def request_model(self, request_data: dict):
        auth = self.model_settings.get_trace_back("auth", {})
        auth = auth if isinstance(auth, dict) else {}
        if "secret_id" not in auth or "secret_key" not in auth:
            raise Exception("[Request] Missing 'secret_id' or 'secret_key' in auth. Use .set_settings('model.Hunyuan.auth', { 'secret_id': '***', 'secret_key': '***' }) to set it.")
        cred = credential.Credential(auth["secret_id"], auth["secret_key"])
        httpProfile = HttpProfile()
        httpProfile.endpoint = auth["endpoint"] if "endpoint" in auth else "hunyuan.tencentcloudapi.com"
        clientProfile = ClientProfile()
        clientProfile.httpProfile = httpProfile
        client = hunyuan_client.HunyuanClient(cred, "", clientProfile)

        request = models.ChatCompletionsRequest()
        request.from_json_string(json.dumps(request_data))

        resp = client.ChatCompletions(request)
        if isinstance(resp, types.GeneratorType):
            for event in resp:
                yield event
        else:
            if self.request.settings.get_trace_back("is_debug"):
                print(f"[Request] Server Response Message: { str(resp) }")

    async def broadcast_response(self, response_generator):
        response_message = {}
        message_content = ""
        async for part in response_generator:
            part_data = json.loads(part["data"])
            try:
                yield({ "event": "response:delta_origin", "data": part_data })
                yield({ "event": "response:delta", "data": part_data["Choices"][0]["Delta"]["Content"] })
                message_content += part_data["Choices"][0]["Delta"]["Content"]
            except Exception as e:
                if self.request.settings.get_trace_back("is_debug"):
                    print(f"[Request] Server Response Message: { str(part_data) }")
        response_message = part_data.copy()
        response_message["Choices"][0].update({
            "Message": {
                "Role": response_message["Choices"][0]["Delta"]["Role"],
                "Content": message_content,
            }
        })
        del response_message["Choices"][0]["Delta"]
        yield({ "event": "response:done_origin", "data": response_message })
        yield({ "event": "response:done", "data": message_content })
    
    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }

def export():
    return("Hunyuan", Hunyuan)



import yaml
from .preset_chunks import preset_chunks

def create_chunk_with_executor(workflow, chunk_id, chunk_info, chunk_executor):
    workflow.chunk(chunk_id=chunk_id, **chunk_info)(chunk_executor)

def start_yaml(workflow, yaml_dict, *, draw):
    # Check Essential Items
    if "chunks" not in yaml_dict or "connections" not in yaml_dict:
        raise Exception("Workflow YAML must include item 'chunks' and item 'connections'.")
    # Create Chunks
    for chunk_id, chunk_info in yaml_dict["chunks"].items():
        chunk_type = None
        chunk_executor_name = None
        created = False
        # Check Chunk Info Format
        if not isinstance(chunk_info, dict):
            raise Exception(f"Content of chunk '{ chunk_id }' should be a dict.")
        # Type Stated
        if "type" in chunk_info:
            chunk_type = chunk_info["type"]
            if chunk_type in preset_chunks["type_list"]:
                preset_chunks["create"][chunk_info["type"]](workflow, chunk_id, chunk_info)
                created = True
        # Executor Stated
        if "executor" in chunk_info:
            chunk_executor_name = chunk_info["executor"]
            chunk_executor = workflow.executor_manager.get(chunk_executor_name)
            if chunk_executor:
                create_chunk_info = chunk_info.copy()
                del create_chunk_info["executor"]
                create_chunk_with_executor(workflow, chunk_id, create_chunk_info, chunk_executor)
                created = True
            else:
                raise Exception(f"Can not find registered chunk executor named '{ chunk_executor_name }'.")
        # Can Not Create Chunk
        if not created:
            raise Exception(f"Can not create chunk: \nChunk type: { chunk_type }\nExecutor name: { chunk_executor_name }")
    # State Connections
    for connection in yaml_dict["connections"]:
        from_chunk_id = None
        to_chunk_id = None
        from_chunk_handle = None
        to_chunk_handle = None
        # full expression
        if isinstance(connection, dict):
            if "from" not in connection:
                raise Exception(f"Must state origin chunk id in key 'from'.\nConnection: { str(connection) }")
            if "to" not in connection:
                raise Exception(f"Must state target chunk id in key 'to'.\nConnection: { str(connection) }")
            from_info_list = connection["from"].split(".")
            from_chunk_id = from_info_list[0]
            from_chunk_handle = from_info_list[1] if len(from_info_list) > 1 else None
            to_info_list = connection["to"].split(".")
            to_chunk_id = to_info_list[0]
            to_chunk_handle = to_info_list[1] if len(to_info_list) > 1 else None
            from_chunk = (
                workflow.chunks[from_chunk_id].handle(from_chunk_handle)
                if from_chunk_handle
                else workflow.chunks[from_chunk_id]
            )
            to_chunk = (
                workflow.chunks[to_chunk_id].handle(to_chunk_handle)
                if to_chunk_handle
                else workflow.chunks[to_chunk_id]
            )
            from_chunk.connect_to(to_chunk)
        # simple expression
        elif isinstance(connection, str):
            connection_info_list = connection.split("->")
            if len(connection_info_list) < 2:
                raise Exception(f"Connect chunk id with at least one '->'.\nConnection: { connection }")
            from_chunk_id = None
            from_chunk_handle = None
            to_chunk_id = None
            to_chunk_handle = None
            is_first_chunk = True
            for connection_chunk in connection_info_list:
                connection_chunk_info_list = connection_chunk.split(".")
                connection_chunk_id = connection_chunk_info_list[0]
                connection_chunk_handle = connection_info_list[1] if len(connection_chunk_info_list) > 1 else None
                if is_first_chunk:
                    from_chunk_id = connection_chunk_id
                    from_chunk_handle = connection_chunk_handle
                    is_first_chunk = False
                else:
                    to_chunk_id = connection_chunk_id
                    to_chunk_handle = connection_chunk_handle
                    from_chunk = (
                        workflow.chunks[from_chunk_id].handle(from_chunk_handle)
                        if from_chunk_handle
                        else workflow.chunks[from_chunk_id]
                    )
                    to_chunk = (
                        workflow.chunks[to_chunk_id].handle(to_chunk_handle)
                        if to_chunk_handle
                        else workflow.chunks[to_chunk_id]
                    )
                    from_chunk.connect_to(to_chunk)
                    from_chunk_id = to_chunk_id
                    from_chunk_handle = to_chunk_handle
    # Start Workflow
    if draw:
        return workflow.draw()
    else:
        return workflow.start()

def start_yaml_from_path(workflow, yaml_path, *, draw):
    try:
        with open(yaml_path, "r", encoding="utf-8") as yaml_file:
            yaml_dict = yaml.safe_load(yaml_file)
            return start_yaml(workflow, yaml_dict, draw=draw)
    except Exception as e:
        raise Exception(f"[Agently Workflow] Error occured when start YAML from path '{ yaml_path }'.\nError: { str(e) }")

def start_yaml_from_str(workflow, yaml_str, *, draw):
    try:
        yaml_dict = yaml.safe_load(yaml_str)
        return start_yaml(workflow, yaml_dict, draw=draw)
    except Exception as e:
        raise Exception(f"[Agently Workflow] Error occured when start YAML from string:\n{ yaml_str }\nError: { str(e) }")


import base64
import sys
from .constants import EXECUTOR_TYPE_LOOP, EXECUTOR_TYPE_CONDITION

def draw_with_mermaid(schema_compiled_data):
    """使用 mermaid 绘制出图形，可点击 https://mermaid-js.github.io/mermaid-live-editor/edit 粘贴查看效果"""

    LOOP_CLASS_NAME = 'loop_style'
    CHUNK_CLASS_NAME = 'chunk_style'
    CONDITION_CLASS_NAME = 'condition_chunk_style'
    # 修复文本展示，避免特殊字符
    def fix_display_text(label: str):
        if not label:
            return label
        escaped_string = label.replace('"', '&quot;')
        return f'"{escaped_string}"'

    def to_node_symbol(chunk, title=None):
        chunk_name = title or fix_display_text(chunk["title"] or f'chunk-{chunk["id"]}' or 'Unknow chunk')
        # 条件节点，渲染成菱形
        if chunk.get('type') == EXECUTOR_TYPE_CONDITION:
            return f"{chunk['id']}" + "{{" + f"{chunk_name}" + "}}:::" + f"{CONDITION_CLASS_NAME}"
        return f"{chunk['id']}({chunk_name}):::{CHUNK_CLASS_NAME}"
    
    helper = { "loop_count": 0 }
    indent_str = '    '
    def generate_loop_body(chunk):
        """生成循环块"""
        loop_body = []
        # 判断 chunk 是否为子workflow，如果是，尝试绘制出来
        loop_info = (chunk.get('extra_info') or {}).get('loop_info') or {}
        if loop_info.get('type') == 'workflow':
            loop_body = generate_body(loop_info.get('detail'))
        else:
            loop_body.append(to_node_symbol(chunk, loop_info.get('detail')))

        # 生成唯一标题
        helper['loop_count'] += 1
        loop_title = f"Loop_{helper['loop_count']}"

        return {
            "title": loop_title,
            "body": [f"subgraph {loop_title}\n{indent_str}direction LR"] + loop_body + ["end"]
        }


    def generate_body(schema_data):
        """主体生成函数"""
        label_split = ' -->-- '
        condition_label_split = ' -- ◇ -- '
        edges = schema_data['edges']
        chunks = schema_data['chunks']
        chunk_map = {}
        for chunk in chunks:
            chunk_map[chunk['id']] = chunk
        graph_partial_list = []
        marked_loop_chunks = [] # 已被标记过的 loop chunk
        graph_loop_blocks = []
        exchanged_graph_symbol = {}
        for idx, edge in enumerate(edges):
            source = chunk_map[edge['source']] if edge.get('source') else None
            target = chunk_map[edge['target']] if edge.get('target') else None
            if source and target:
                """连接起点和终点"""
                # 处理条件判断
                condition_symbol = None
                if edge.get('condition'):
                    condition_symbol = f"condition{idx}" + "{?}"

                # 处理连接手柄
                start_label_symbol = ''
                end_label_symbol = ''
                if edge.get('source_handle') and edge.get('target_handle'):
                    start_label_symbol = edge.get('source_handle')
                    end_label_symbol = edge.get('target_handle')

                # 判断是否是loop
                if target.get('type') == EXECUTOR_TYPE_LOOP:
                    # 如果未被处理过，则生成一个loop区域
                    if edge.get('target') not in marked_loop_chunks:
                        # 生成绘图数据
                        loop_graph_info = generate_loop_body(target)
                        # 加入节点
                        graph_loop_blocks.extend(loop_graph_info['body'])
                        # 标识已处理过
                        marked_loop_chunks.append(edge.get('target'))
                        # 登记该loop 节点信息，后续全重置连接到该循环节点
                        exchanged_graph_symbol[edge['target']] = loop_graph_info['title']
                # 处理本次连接标识
                full_labels_symbol = fix_display_text(
                    f"{start_label_symbol}{condition_label_split if condition_symbol else label_split}{end_label_symbol}")
                # 处理最终的起始节点
                latest_target_label = f"{exchanged_graph_symbol.get(edge['target'])}:::{LOOP_CLASS_NAME}" if exchanged_graph_symbol.get(
                    edge['target']) else to_node_symbol(target)
                latest_source_label = f"{exchanged_graph_symbol.get(edge['source'])}:::{LOOP_CLASS_NAME}" if exchanged_graph_symbol.get(
                    edge['source']) else to_node_symbol(source)
                # 加入绘制图中
                graph_partial_list.append(
                    f"{latest_source_label} -.-> |{full_labels_symbol}| {latest_target_label}")
        return graph_loop_blocks + graph_partial_list

    class_str = '\n'.join([
        f"classDef {CHUNK_CLASS_NAME} fill:#fbfcdb,stroke:#666,stroke-width:1px,color:#333;",
        f"classDef {CONDITION_CLASS_NAME} fill:#ECECFF,stroke:#9370DB,stroke-width:1px,color:#333;",
        f"classDef {LOOP_CLASS_NAME} fill:#f5f7fa,stroke:#666,stroke-width:1px,color:#333,stroke-dasharray: 5 5",
    ])
    return (
        "%%{ init: { 'flowchart': { 'curve': 'linear' }, 'theme': 'neutral' } }%%\n" +
        "%% Rendered By Agently %%\nflowchart LR\n" + class_str + "\n" +
        '\n'.join(map(lambda part: indent_str + part, generate_body(schema_compiled_data)))
    )

def draw_image_in_jupyter(schema_compiled_data):
    if 'IPython' in sys.modules:
        from IPython.display import Image, display
        mm_code = draw_with_mermaid(schema_compiled_data)
        graphbytes = mm_code.encode("utf8")
        base64_bytes = base64.b64encode(graphbytes)
        base64_string = base64_bytes.decode("ascii")
        display(Image(url="https://mermaid.ink/img/" + base64_string))
    else:
        print('Please use within the Jupyter environment.')



import os
import json
import sqlite3
from .utils import StorageABC

class SQLite(StorageABC):
    def __init__(self, db_name: str="default", settings: object={}):
        self.settings = settings
        self.path = self.settings.get("storage.SQLite.path") or None
        self.file_name = self.settings.get("storage.SQLite.file_name") or ".Agently.db"
        if self.path and not self.path.endswith("/"):
            self.path = self.path + "/"
        if not self.file_name.endswith(".db"):
            self.file_name = self.file_name + ".db"
        self.db = f"{ self.path }{ self.file_name }" if self.path else self.file_name
        self.space_name = db_name
        self.conn = None
        self.cursor = None

    def __connect(self):
        self.conn = sqlite3.connect(self.db)
        self.cursor = self.conn.cursor()

    def __connect_if_exists(self):
        if os.path.exists(self.db):
            self.conn = sqlite3.connect(self.db)
            self.cursor = self.conn.cursor()
            return True
        else:
            return False
    
    def __close(self):
        if self.conn:
            self.cursor.close()
            self.conn.close()
            self.conn = None
            self.cursor = None

    def __commit_and_close(self):
        if self.conn:
            self.conn.commit()
            self.cursor.close()
            self.conn.close()
            self.conn = None
            self.cursor = None

    def __create_table_if_not_exists(self, table_name: str):
        self.cursor.execute(f"CREATE TABLE IF NOT EXISTS `{ self.space_name }_{ table_name }` (key TEXT PRIMARY KEY, value TEXT)")
    
    def __drop_table_if_exists(self, table_name: str):
        self.cursor.execute(f"DROP TABLE IF EXISTS `{ self.space_name }_{ table_name }`")
    
    def set(self, table_name: str, key: str, value: any):
        self.__connect()
        self.__create_table_if_not_exists(table_name)
        self.cursor.execute(
f"""INSERT INTO `{ self.space_name }_{ table_name }` (`key`, `value`)
    VALUES (?, ?)
    ON CONFLICT(`key`)
    DO UPDATE SET `key`=excluded.key, `value`=excluded.value
""",
            (key, json.dumps(value))
        )
        self.__commit_and_close()
        return self

    def set_all(self, table_name:str, full_data: dict):
        self.__connect()
        self.__drop_table_if_exists(table_name)
        self.__create_table_if_not_exists(table_name)
        for key, value in full_data.items():
            self.set(table_name, key, value)
        self.__commit_and_close()
        return self

    def remove(self, table_name: str, key: str):
        self.__connect()
        self.__create_table_if_not_exists(table_name)
        self.cursor.execute(f"DELETE FROM `{ self.space_name }_{ table_name }` WHERE `key` = ?", (key,))
        self.__commit_and_close()
        return self

    def update(self, table_name:str, update_data: dict):
        self.__connect()
        self.__create_table_if_not_exists(table_name)
        for key, value in update_data.items():
            self.set(table_name, key, value)
        self.__commit_and_close()
        return self

    def get(self, table_name: str, key: str):
        if self.__connect_if_exists():
            try:
                self.cursor.execute(f"SELECT `value` FROM `{ self.space_name }_{ table_name }` WHERE `key` = ?", (key,))
                result = self.cursor.fetchone()
            except sqlite3.OperationalError as e:
                result = None
            self.__close()
            if result:
                return json.loads(result[0])
            else:
                return None
        else:
            return None

    def get_all(self, table_name: str, keys: (list, None)=None):
        if self.__connect_if_exists():
            if keys:            
                result = {}
                for key in keys:
                    value = self.get(table_name, key)
                    if value:
                        result.update({ key: value })
                    else:
                        result.update({ key: None })
                self.__close()
                return result
            else:
                table_data = {}
                try:
                    self.cursor.execute(f"SELECT `key`, `value` FROM `{ self.space_name }_{ table_name }`")
                    results = self.cursor.fetchall()
                except sqlite3.OperationalError as e:
                    results = []
                for row in results:
                    key, value = row[0], json.loads(row[1])
                    table_data.update({ key: value })
                self.__close()
                return table_data
        else:
            return {}

def export():
    return ("SQLite", SQLite)


from .utils import FacilityABC
from Agently.utils import RuntimeCtx, RuntimeCtxNamespace

class Embedding(FacilityABC):
    def __init__(self, *, storage: object, plugin_manager: object, settings: object):
        self.plugin_manager = plugin_manager
        self.settings = settings

    def OpenAI(self, text: str, options: dict={}):
        from openai import OpenAI
        import httpx
        settings = RuntimeCtxNamespace("embedding.OpenAI", self.settings)
        # Prepare Client Params
        client_params = {}
        base_url = settings.get_trace_back("url")
        if base_url:
            client_params.update({ "base_url": base_url })
        proxy = self.settings.get_trace_back("proxy")
        if proxy:
            client_params.update({ "http_client": httpx.Client( proxies = proxy ) })
        api_key = settings.get_trace_back("auth.api_key")
        if api_key:
            client_params.update({ "api_key": api_key })
        else:
            raise Exception("[Embedding] OpenAI require api_key. use Agently.facility.set_settings('embedding.OpenAI.auth', { 'api_key': '<Your-API-Key>' }) to set it.")
        if "model" not in options:
            options["model"] = "text-embedding-ada-002"
        # Create Client
        client = OpenAI(**client_params)
        # Request
        result = client.embeddings.create(
            input = text.replace("\n", " "),
            **options
        )
        if result.data:
            return result.data[0].embedding
        else:
            raise Exception(f"[Embedding] OpenAI Error: { dict(result) }")

    def ERNIE(self, text_list: (str, list), options: dict={}):
        import erniebot
        settings = RuntimeCtxNamespace("embedding.ERNIE", self.settings)
        erniebot.api_type = "aistudio"
        access_token = settings.get_trace_back("auth.aistudio")
        if access_token:
            erniebot.access_token = access_token
        else:
            raise Exception("[Embedding] ERNIE require aistudio access token. use Agently.facility.set_settings('embedding.ERNIE.auth', { 'aistudio': '<Your-Access-Token>' }) to set it.")
        if "model" not in options:
            options["model"] = "ernie-text-embedding"
        if not isinstance(text_list, list):
            text_list = [text_list]
        result = erniebot.Embedding.create(
            input = text_list,
            **options
        )
        result = dict(result)
        if result["rcode"] == 200:
            return result["rbody"]["data"]
        else:
            raise Exception(f"[Embedding] ERNIE Error: { result }")

    def ZhipuAI(self, text: str, options: dict={}):
        import zhipuai
        settings = RuntimeCtxNamespace("embedding.ZhipuAI", self.settings)
        api_key = settings.get_trace_back("auth.api_key")
        if api_key:
            zhipuai.api_key = api_key
        else:
            raise Exception("[Embedding] ZhipuAI require api_key. use Agently.facility.set_settings('embedding.ZhipuAI.auth', { 'api_key': '<Your-API-Key>' }) to set it.")
        if "model" not in options:
            options["model"] = "text_embedding"
        result = zhipuai.model_api.invoke(
            prompt = text,
            **options
        )
        if "data" in result:
            return result["data"]["embedding"]
        else:
            raise Exception(f"[Embedding] ZhipuAI Error: { result }")

    def Google(self, text: str, options: dict={}):
        import json
        import httpx
        settings = RuntimeCtxNamespace("embedding.Google", self.settings)
        api_key = settings.get_trace_back("auth.api_key")
        if api_key == None:
            raise Exception("[Embedding] Google require api_key. use Agently.facility.set_settings('embedding.google.auth', { 'api_key': '<Your-API-Key>' }) to set it.")
        proxy = self.settings.get_trace_back("proxy")
        model = "embedding-001"
        if "model" in options:
            model = options["model"]
            del options["model"]
        request_params = {
            "data": json.dumps({
                "content": {
                    "parts": [{ "text": text }]
                },
                **options
            }),
            "timeout": None,
        }
        result = httpx.post(
            f"https://generativelanguage.googleapis.com/v1/models/{ model }:embedContent?key={ api_key }",
            **request_params
        )
        if result.status_code:
            return result.json()["embedding"]["values"]
        else:
            raise Exception(f"[Embedding] Google Error: { dict(result) }")

def export():
    return ("embedding", Embedding)


from ...lib.Store import Store

def condition_executor(inputs, store: Store, **sys_info):
    """条件执行器，会从其 extra_info['condition_info'] 中取到判断条件和结果符号，最终返回命中判断的结果符号，同时透传上游的值"""
    chunk_data = sys_info['chunk']['data']
    condition_info = (chunk_data.get('extra_info') or {}).get('condition_info') or {}
    normal_conditions = condition_info.get('conditions')
    # 从众多条件中依次判断，如果命中了，则返回对应条件的结果 id
    if normal_conditions and len(normal_conditions) > 0:
        for condition_desc in normal_conditions:
            condition = condition_desc.get('condition')
            if condition and condition(inputs['default'], store):
                return {
                    "condition_signal": condition_desc.get('condition_signal'),
                    "values": inputs['default']
                }
    # 否则返回 else 的结果 id
    else_condition = condition_info.get('else_condition')
    return {
        "condition_signal": else_condition.get('condition_signal'),
        "values": inputs['default']
    }



def create(workflow, chunk_id, chunk_info):
    @workflow.chunk(chunk_id=chunk_id, **chunk_info)
    def user_input_executor(inputs, storage):
        input_str = input(
            chunk_info["placeholder"] if "placeholder" in chunk_info else "[User]:",
        )
        return input_str


import os
import sys
import json
import importlib
from Agently.utils import find_json

def install_plugins(plugin_manager, tool_manager, settings):
    dir_path = os.path.dirname(os.path.abspath(__file__))
    for item in os.listdir(dir_path):
        if os.path.exists(f"{ dir_path }/{ item }/__init__.py"):
            module_plugins = importlib.import_module(f".{ item }", package = __package__)
            if not hasattr(module_plugins, "export"):
                raise Exception(f"[Plugin Manager] Function 'export' must be stated in '__init__.py' file in module plugin dir: { dir_path }/{ item }")
            module_plugin_export = getattr(module_plugins, "export")()
            module_plugin_list = module_plugin_export[0]
            module_default_settings = module_plugin_export[1]
            orders = module_plugin_export[2] if len(module_plugin_export) > 2 else None
            try:
                for plugin_info in module_plugin_list:
                    plugin_manager.register(plugin_info[0], plugin_info[1], plugin_info[2])
            except:
                raise Exception(f"[Plugin Manager] Function 'export' in '__init__.py' must return a list of tuple (<module name>, <plugin name>, <plugin method>).")
            if module_default_settings:
                #plugin_manager.update_settings(module_default_settings)
                for key, value in module_default_settings.items():
                    if isinstance(value, str) and value.lower() == "true":
                        module_default_settings[key] = True
                    if isinstance(value, str) and value.lower() == "false":
                        module_default_settings[key] = False
                    json_string = find_json(value)
                    if json_string != None and json_string != '':
                        module_default_settings[key] = json.loads(json_string)
                    settings.set(key, module_default_settings[key])
            if orders and item == "agent_component":
                agent_component_list = plugin_manager.get_agent_component_list()
                for process_type in orders:
                    order_list = orders[process_type].split(",")
                    current_mode = "firstly"
                    order_settings = {
                        "firstly": [],
                        "normally": [],
                        "finally": [],
                    }
                    for agent_component_name in order_list:
                        agent_component_name = agent_component_name.replace(" ", "")
                        if agent_component_name == "...":
                            current_mode = "finally"
                        elif agent_component_name in agent_component_list:
                            order_settings[current_mode].append(agent_component_name)
                    ordered_agent_component_list = order_settings["firstly"].copy()
                    ordered_agent_component_list.extend(order_settings["finally"].copy())
                    for agent_component_name in agent_component_list:
                        if agent_component_name not in ordered_agent_component_list:
                            order_settings["normally"].append(agent_component_name)
                    settings.set(f"plugin_settings.agent_component.orders.{ process_type }", order_settings)
    for tool_package_name, ToolPackageClass in plugin_manager.plugins_runtime_ctx.get("tool", {}).items():
        tool_package = ToolPackageClass(tool_manager)
        for tool_name, tool_info in tool_package.export().items():
            if "desc" not in tool_info:
                raise Exception(f"[Plugin] Tool '{ tool_name }' in tool package '{ tool_package_name }' must export key 'desc'.")
            if "args" not in tool_info or not isinstance(tool_info["args"], dict):
                raise Exception(f"[Plugin] Tool '{ tool_name }' in tool package '{ tool_package_name }' must export key 'args', set the value as {{}} if need no args.")
            if "func" not in tool_info or not callable(tool_info["func"]):
                raise Exception(f"[Plugin] Tool '{ tool_name }' in tool package '{ tool_package_name }' must export key 'func'.")
            tool_manager.register(
                tool_name = tool_name,
                desc = tool_info["desc"],
                args = tool_info["args"],
                func = tool_info["func"],
                require_proxy = tool_info["require_proxy"] if "require_proxy" in tool_info else False,
                categories = tool_info["categories"] if "categories" in tool_info else None
            )

def install(Agently):
    install_plugins(Agently.global_plugin_manager, Agently.global_tool_manager, Agently.global_settings)
    Agently.facility.refresh_plugins()

def install_to_agent(agent):
    install_plugins(agent.plugin_manager, agent.tool_manager, agent.settings)
    agent.refresh_plugins()


from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, find_json, format_request_messages
from Agently.utils import RuntimeCtxNamespace
import httpx
import json

class Claude(RequestABC):
    def __init__(self, request):
        self.request = request
        self.model_name = "Claude"
        self.model_settings = RuntimeCtxNamespace(f"model.{ self.model_name }", self.request.settings)
        if not self.model_settings.get_trace_back("message_rules.no_multi_system_messages"):
            self.model_settings.set("message_rules.no_multi_system_messages", False)
        if not self.model_settings.get_trace_back("message_rules.strict_orders"):
            self.model_settings.set("message_rules.strict_orders", True)
        if not self.model_settings.get_trace_back("message_rules.no_multi_type_messages"):
            self.model_settings.set("message_rules.no_multi_type_messages", False)
        self.default_options = {
            "model": "claude-3-sonnet-20240229",
            "max_tokens": 4096,
            "stream": True,
        }
    
    def construct_request_messages(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[GENERAL INSTRUCTION]\n{ to_instruction(general_instruction_data) }" }] })
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[ROLE SETTINGS]\n{ to_instruction(role_data) }" }] })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[ABOUT USER]\n{ to_instruction(user_info_data) }" }] })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "assistant", "content": [{"type": "text", "text": to_instruction(headline_data) }] })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        # - files
        files_data = self.request.request_runtime_ctx.get_trace_back("prompt.files")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        prompt_text = ""
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            prompt_text = to_instruction(prompt_input_data)
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[HELPFUL INFORMATION]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[OUTPUT REQUIREMENT]"] = {
                        "TYPE": "JSON can be parsed in Python",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[OUTPUT REQUIERMENT]"] = str(prompt_output_data)
            prompt_text = to_prompt_structure(prompt_dict, end="[OUTPUT]:\n")
        request_messages.append({ "role": "user", "content": [{"type": "text", "text": prompt_text }] })
        return request_messages

    

    def generate_request_data(self):
        options = self.model_settings.get_trace_back("options", {})
        return {
            "messages": format_request_messages(self.construct_request_messages(), self.model_settings),
            "options": options,
        }

    async def request_model(self, request_data: dict):
        api_key = self.model_settings.get_trace_back("auth.api_key")
        base_url = self.model_settings.get_trace_back("url", "https://api.anthropic.com/v1")
        if base_url.endswith("/"):
            base_url = base_url[:-1]
        proxy = self.request.settings.get_trace_back("proxy")
        messages = request_data["messages"]
        system_prompt = ""
        request_messages = []
        for message in messages:
            if message["role"] == "system":
                system_prompt += f"{ message['content'][0]['text'] }\n"
            else:
                request_messages.append(message)
        options = request_data["options"]
        if system_prompt != "":
            options.update({ "system": system_prompt })
        for key, value in self.default_options.items():
            if key not in options:
                options.update({ key: value })
        request_params = {
            "headers": {
                "x-api-key": api_key,
                "anthropic-version": "2023-06-01",
                "content-type": "application/json",
            },
            "data": json.dumps({ "messages": request_messages, **options }),
            "timeout": None,
        }
        client_params = {}
        if proxy:
            client_params["proxy"] = proxy
        async with httpx.AsyncClient(**client_params) as client:
            async with client.stream(
                "POST",
                f"{ base_url }/messages",
                **request_params
            ) as response:
                async for chunk in response.aiter_lines():
                    yield chunk

    async def broadcast_response(self, response_generator):
        full_message = {}
        buffer = ""
        async for chunk in response_generator:
            if chunk.startswith("data: "):
                content = json.loads(chunk[6:])
                yield({ "event": "response:delta_origin", "data": content })
                if content["type"] == "message_start":
                    full_message.update(content["message"])
                if content["type"] == "content_block_delta":
                    delta = content["delta"]["text"]
                    if delta is not None:
                        yield({ "event": "response:delta", "data": delta })
                        buffer += delta
                if content["type"] == "message_delta":
                    full_message.update(content["delta"])
                if content["type"] == "message_stop":
                    full_message["content"].append({ "type": "text", "text": buffer })
                    yield({ "event": "response:done_origin", "data": full_message })
                    yield({ "event": "response:done", "data": buffer })

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }

def export():
    return ("Claude", Claude)


# 入口类型
EXECUTOR_TYPE_START = 'Start'
# 结束类型
EXECUTOR_TYPE_END = 'End'

# 条件类型
EXECUTOR_TYPE_CONDITION = 'Condition'

# 内置声明好的执行器
BUILT_IN_EXECUTOR_TYPES = [
  EXECUTOR_TYPE_START,
  EXECUTOR_TYPE_END,
  EXECUTOR_TYPE_CONDITION
]
# 循环类型
EXECUTOR_TYPE_LOOP = 'Loop'

# 常规类型
EXECUTOR_TYPE_NORMAL = 'Normal'

SYS_SYMBOL = 'WORKFLOW_SYS_DATA'

# workflow 初始数据的句柄名
WORKFLOW_START_DATA_HANDLE_NAME = f'{SYS_SYMBOL}:{EXECUTOR_TYPE_START}'
WORKFLOW_END_DATA_HANDLE_NAME = f'{SYS_SYMBOL}:{EXECUTOR_TYPE_END}'

# 默认的输入 handle
DEFAULT_INPUT_HANDLE_VALUE = 'default'
DEFAULT_INPUT_HANDLE = {'handle': DEFAULT_INPUT_HANDLE_VALUE}
# 默认的输出 handle，表示数据输出的总出口（会输出所有数据）
DEFAULT_OUTPUT_HANDLE_VALUE = '*'
DEFAULT_OUTPUT_HANDLE = {'handle': DEFAULT_OUTPUT_HANDLE_VALUE}



import asyncio
from .RuntimeCtx import RuntimeCtx, RuntimeCtxNamespace

class ToolManager(object):
    def __init__(self, *, parent: object=None):
        self.parent = parent
        self.tool_runtime_ctx = RuntimeCtx(parent = self.parent.tool_runtime_ctx if self.parent else None)
        self.category_dict = {}
        self.tool_proxy = None
        self.runtime_cache = {}

    def set_tool_proxy(self, proxy: str):
        self.tool_proxy = proxy
        return self

    def get_tool_proxy(self):
        if self.tool_proxy:
            return self.tool_proxy
        elif self.parent:
            return self.parent.get_tool_proxy()
        else:
            return None

    def register(self, tool_name: str, desc: any, args: dict, func: callable, *, categories: (str, list) = None, require_proxy: bool=False):
        tool_info = {
            "tool_name": tool_name,
            "desc": desc,
            "args": args,
            "func": func,
            "require_proxy": require_proxy
        }
        if categories:
            if isinstance(categories, str):
                categories = [categories]
            tool_info.update({ "categories": categories })
            for category in categories:
                if category not in self.category_dict:
                    self.category_dict.update({ category: None })
        self.tool_runtime_ctx.update(tool_name, tool_info)
        return self

    def tool(self, tool_name: str):
        return RuntimeCtxNamespace(tool_name, self.tool_runtime_ctx)

    def get_tool_info(self, tool_name: str, *, with_args: bool=False, full: bool=False):
        full_info = self.tool(tool_name).get_trace_back()
        if full_info:
            if full:
                return full_info
            else:
                return {
                    "tool_name": full_info["tool_name"] if "tool_name" in full_info else tool_name,
                    "desc": full_info["desc"] if "desc" in full_info else "",
                    "args": full_info["args"] if "args" in full_info else {},
                } if with_args else {
                    "tool_name": full_info["tool_name"] if "tool_name" in full_info else tool_name,
                    "desc": full_info["desc"] if "desc" in full_info else "",
                }
        else:
            return None

    def get_tool_func(self, tool_name: str):
        full_info = self.tool(tool_name).get_trace_back()
        if full_info:
            if "func" in full_info and callable(full_info["func"]):
                return full_info["func"]
            else:
                return None
        else:
            return None

    def call_tool_func(self, tool_name: str, *args, **kwargs):
        func = self.get_tool_func(tool_name)
        if func:
            if asyncio.iscoroutinefunction(func):
                loop = asyncio.get_event_loop()
                return loop.run_until_complete(func(*args, **kwargs))
            else:
                return func(*args, **kwargs)
        else:
            return None

    def get_tool_dict(self, *, categories: (str, list)=None, with_args=False):
        full_tool_dict = self.tool_runtime_ctx.get_trace_back()
        result = {}
        if categories:
            if isinstance(categories, str):
                categories = [categories]
            for tool_name, tool_info in full_tool_dict.items():
                if "categories" in tool_info:
                    for category in categories:
                        if category in tool_info["categories"]:
                            result.update({ tool_name: self.get_tool_info(tool_name, with_args=with_args) })
        else:
            for tool_name in full_tool_dict.keys():
                result.update({ tool_name: self.get_tool_info(tool_name, with_args=with_args) })
        return result

    def get_tool_list(self, *, categories: (str, list)=None, with_args=False):
        full_tool_dict = self.tool_runtime_ctx.get_trace_back()
        result = []
        if categories:
            if isinstance(categories, str):
                categories = [categories]
            for tool_name, tool_info in full_tool_dict.items():
                if "categories" in tool_info:
                    for category in categories:
                        if category in tool_info["categories"]:
                            result.append(self.get_tool_info(tool_name, with_args=with_args))
        else:
            for tool_name in full_tool_dict.keys():
                result.append(self.get_tool_info(tool_name, with_args=with_args))
        return result

    def set_category_desc(self, category: str, desc: any):
        self.category_dict.update({ category: desc })
        return self

    def get_category_dict(self):
        result = {}
        if self.parent:
            result = self.parent.get_category_dict()
        for key, value in self.category_dict.items():
            result.update({ key: value })
        return result


from ...lib.Store import Store
from ...lib.constants import WORKFLOW_END_DATA_HANDLE_NAME

def end_executor(inputs, store: Store, **sys_info):
    sys_info['sys_store'].set(WORKFLOW_END_DATA_HANDLE_NAME, inputs)
    return inputs


import uuid
import copy
from typing import TYPE_CHECKING, TypeVar, Type
from ..utils.chunk_helper import deep_copy_simply
from ..utils.find import has_target_by_attr
from ..lib.constants import EXECUTOR_TYPE_NORMAL, BUILT_IN_EXECUTOR_TYPES, DEFAULT_OUTPUT_HANDLE_VALUE, DEFAULT_INPUT_HANDLE_VALUE
from .Condition import ConditionAbility
from .Loop import LoopAbility
from .helper import exposed_interface

if TYPE_CHECKING:
    from ..Schema import Schema

T = TypeVar('T')


class SchemaChunk(ConditionAbility, LoopAbility):
    """
    Workflow Schema 中的单个 chunk 结构，用于辅助提供操作的 API，实际运行时，提供 chunk 描述参数交给 MainExecutor 执行。
    注意，仅当connect_to/loop 调用时，才会实质创建运行时相关参数，否则仅仅是参数挂载
    """

    def __init__(self, workflow_schema: 'Schema' = None, type=EXECUTOR_TYPE_NORMAL, executor: callable = None, **chunk_desc):
        """
        必选参数 executor，或 type 为 'Start'
        """
        # 校验必填字段(要么 type 为特殊类型，否则必须包含 executor)
        if not type or (type not in BUILT_IN_EXECUTOR_TYPES):
            if not executor:
                raise ValueError("Missing required key: 'executor'")

        self.workflow_schema = workflow_schema

        # chunk 运行时参数，每个字段均参与运行过程
        self.chunk = {
            'id': chunk_desc.get('id', str(uuid.uuid4())),
            'title': chunk_desc.get('title'),
            'type': type or EXECUTOR_TYPE_NORMAL,
            'executor': executor,
            'handles': self._fix_handles(chunk_desc.get('handles')),
            # 连接条件（默认无条件连通）
            'connect_condition': chunk_desc.get('connect_condition') or None,
            # 连接条件的描述（if/elif，condition_id 信息等）
            'connect_condition_detail': chunk_desc.get('connect_condition_detail') or None,
            # 扩展信息
            'extra_info': chunk_desc.get('extra_info') or {}
        }

        self._pre_command_interceptors = []
        self._post_command_interceptors = []
        # 支撑参数空间
        self.shared_ns = {
            'connection_ns': {
                'active_handle': chunk_desc.get('active_handle', None)
            }
        }
        # 链路信息空间（会在chunk流转到另一个chunk时透传）
        self.shared_chain_ns = {}
        super().__init__()
    
    @exposed_interface(type='attr')
    def handle(self, name: str) -> 'SchemaChunk':
        """Chunk 的连接点，注意 handle 可能在此时还未注册，在其进行连接操作时，会自动帮其注册"""
        # 创建影子 chunk，逻辑分支继续延续
        shadow_chunk = self.copy_shadow_chunk()
        shadow_chunk.set_active_handle(name)
        return shadow_chunk

    @exposed_interface(type='connect_command')
    def connect_to(self, chunk: 'SchemaChunk') -> 'SchemaChunk':
        """
        连接到指定个 Chunk 节点，支持传入 create_chunk 创建好的实例、chunk dict、已定义好的 chunk 名称
        """
        if isinstance(chunk, str):
            chunks = self.workflow_schema.workflow.chunks
            chunk_info_list = chunk.split(".")
            if len(chunk_info_list) < 2:
                chunk_id = chunk_info_list[0]
                if chunk_id not in chunks:
                    raise Exception(
                        f"Can not find '{ chunk_id }' in workflow.chunks.")
                return self._raw_connect_to(chunks[chunk_id])
            else:
                chunk_id = chunk_info_list[0]
                handle_name = chunk_info_list[1]
                if chunk_id not in chunks:
                    raise Exception(
                        f"Can not find '{ chunk_id }' in workflow.chunks.")
                return self._raw_connect_to(chunks[chunk_id].handle(handle_name))
        elif callable(chunk):
            temp_chunk = self.workflow_schema.create_chunk(
                executor=chunk,
                type=EXECUTOR_TYPE_NORMAL,
                title=(
                    f"@{ chunk.__name__ }"
                    if not chunk.__name__ == "<lambda>"
                    else f"lambda-{ str(uuid.uuid4()) }"
                ),
            )
            return self._raw_connect_to(temp_chunk)
        else:
            return self._raw_connect_to(chunk)

    def set_active_handle(self, name: str = None):
        """设置激活的连接点，连接时会默认连接到该点上"""
        self.shared_ns['connection_ns']['active_handle'] = name

    def has_handle(self, name: str, from_inputs=True) -> bool:
        """判断连接手柄是否存在"""
        search_range = self.chunk.get('handles').get(
            'inputs' if from_inputs else 'outputs', [])
        return has_target_by_attr(search_range, 'handle', name)

    def copy_shadow_chunk(self, persist_supports_data=True):
        """获取到当前 chunk 的拷贝份（不干扰 chunk 自身运算逻辑），persist_supports_data 为是否保留支撑参数"""
        shadow_chunk = SchemaChunk(
            workflow_schema=self.workflow_schema,
            **self.get_raw_schema()
        )
        if persist_supports_data:
            shadow_chunk.shared_ns = deep_copy_simply(self.shared_ns)
            shadow_chunk.shared_chain_ns = deep_copy_simply(self.shared_chain_ns)

        return shadow_chunk
    
    def create_rel_chunk(self, persist_supports_data=True, **params) -> 'SchemaChunk':
        """在当前 schema 上创建一个新的 chunk """
        rel_chunk = self.workflow_schema.create_chunk(**params)
        if persist_supports_data:
            # 新挂载的 chunk，仅保留调用链，其它 api 运行数据保留
            rel_chunk.shared_chain_ns = deep_copy_simply(self.shared_chain_ns)
        return rel_chunk

    def get_raw_schema(self, use_origin=False):
        """获取原始的配置"""
        return self.chunk if use_origin else copy.deepcopy(self.chunk)
    
    def _raw_connect_to(self, chunk: 'SchemaChunk') -> 'SchemaChunk':
        """内部原生连接方法"""
        # 如果传入的是一个 dict 类型，自动进行实例化再连接
        target_chunk = self.workflow_schema.create_chunk(
            **chunk) if isinstance(chunk, dict) else chunk
        if type(target_chunk) != type(self):
            raise ValueError(
                f"The 'chunk' parameter to 'connect_to' is invalid.")

        # 检测 chunk 是否在当前 workflow 中有注册
        if not self.workflow_schema.get_chunk(target_chunk.chunk['id']):
            raise ValueError(
                f"The two chunk are not mounted under the same 'workflow'.")

        # 连接两 chunk
        self.workflow_schema.connect_chunk(
            self.chunk['id'],
            target_chunk.chunk['id'],
            self.shared_ns['connection_ns']['active_handle'] or DEFAULT_OUTPUT_HANDLE_VALUE,
            target_chunk.shared_ns['connection_ns']['active_handle'] or DEFAULT_INPUT_HANDLE_VALUE,
            self.chunk.get('connect_condition') or None,
            self.chunk.get('connect_condition_detail') or None,
        )
        target_chunk_shadow = target_chunk.copy_shadow_chunk(persist_supports_data=False)
        # 继承链路信息
        target_chunk_shadow.shared_chain_ns = deep_copy_simply(self.shared_chain_ns)
        return target_chunk_shadow

    def _fix_handles(self, custom_handles):
        """修正用户传入的 handles 配置格式"""
        # 处理默认 handle
        handles = custom_handles or {'inputs': [], 'outputs': []}
        if 'inputs' not in handles:
            handles['inputs'] = []
        if 'outputs' not in handles:
            handles['outputs'] = []

        # 处理 handle 点简写的情况（如：{ "inputs": ['topic', 'type'], "outputs": [{"handle": "output"}] } 中的 inputs 的定义方式）
        def fix_handle_core(handle_list):
            if not handle_list:
                return
            if handle_list and len(handle_list):
                for idx, handle in enumerate(handle_list):
                    if isinstance(handle, str):
                        handle_list[idx] = {"handle": handle}

        fix_handle_core(handles['inputs'])
        fix_handle_core(handles['outputs'])
        return handles



from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace

class UserInfo(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.user_info_runtime_ctx = RuntimeCtxNamespace("user_info", self.agent.agent_runtime_ctx)
        self.user_info_storage = self.agent.global_storage.table("user_info")

    def set_name(self, name: str):
        self.user_info_runtime_ctx = self.__get_user_info_runtime_ctx(target)
        self.user_info_runtime_ctx.set("NAME", name)
        return self.agent

    def set(self, key: any, value: any=None):
        if value is not None:
            self.user_info_runtime_ctx.set(key, value)
        else:
            self.user_info_runtime_ctx.set("DESC", key)
        return self.agent

    def update(self, key: any, value: any=None):
        if value is not None:
            self.user_info_runtime_ctx.update(key, value)
        else:
            self.user_info_runtime_ctx.update("DESC", key)
        return self.agent        

    def append(self, key: any, value: any=None):
        if value is not None:
            self.user_info_runtime_ctx.append(key, value)
        else:
            self.user_info_runtime_ctx.append("DESC", key)
        return self.agent

    def extend(self, key: any, value: any=None):
        if value is not None:
            self.user_info_runtime_ctx.extend(key, value)
        else:
            self.user_info_runtime_ctx.extend("DESC", key)
        return self.agent

    def save(self, role_name: str=None):
        if user_info_name == None:
            user_info_name = self.user_info_runtime_ctx.get("NAME")
        if user_info_name != None and user_info_name != "":
            self.user_info_storage\
                .set(user_info_name, self.user_info_runtime_ctx.get())\
                .save()
            return self.agent
        else:
            raise Exception("[Agent Component: UserInfo] UserInfo attr 'NAME' must be stated before save. Use .set_user_name() to specific that.")

    def load(self, role_name: str):
        user_info_data = self.user_info_storage.get(role_name)
        for key, value in user_info_data.items():
            self.user_info_runtime_ctx.update(key, value)
        return self.agent

    def _prefix(self):
        return {
            "user_info": self.user_info_runtime_ctx.get(),
        }

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "set_user_name": { "func": self.set_name },
                "set_user_info": { "func": self.set },
                "update_user_info": { "func": self.update },
                "append_user_info": { "func": self.append },
                "extend_user_info": { "func": self.extend },
                "save_user_info": { "func": self.save },
                "load_user_info": { "func": self.load },
            },
        }

def export():
    return ("UserInfo", UserInfo)